import numpy as np
import pandas as pd
import statsmodels.api as sm
from patsy import dmatrices
from scipy import optimize
from tqdm import tqdm
from .utils import parse_formula, estimate_misclassification_rates
from .plotting import plot_coefficient_comparison, plot_bootstrap_distributions

class ValidationCorrection:
    def __init__(self):
        self.misclassification_rates = None
        self.formula = None
        self.data = None
        self.val_data = None
        self.y_var = None
        self.x_vars = None
        self.mismeasured_var = None
        self.true_var = None
        self.is_outcome_mismeasured = False
        self.bootstrap_results = None

    def _prepare_data(self, formula, data, val_data=None):
        """Parse formula and prepare data for analysis."""
        parsed = parse_formula(formula)
        self.formula = formula
        self.data = data.copy()
        self.val_data = val_data.copy() if val_data is not None else None
        
        self.y_var = parsed['y_var']
        self.x_vars = parsed['x_vars']
        self.mismeasured_var = parsed['mismeasured_var']
        self.true_var = parsed['true_var']
        self.is_outcome_mismeasured = parsed['is_outcome_mismeasured']

        if val_data is not None:
            self.misclassification_rates = estimate_misclassification_rates(
                val_data[self.mismeasured_var], 
                val_data[self.true_var]
            )

    def _estimate_classification_probs(self, mismeasured, true):
        """
        Estimate classification probabilities θ_{w|x} = P(W=w|X=x)
        from validation data.
        """
        categories = np.unique(np.concatenate([mismeasured, true]))
        n_categories = len(categories)
        theta = np.zeros((n_categories, n_categories))
        
        # Calculate conditional probabilities P(W=w|X=x)
        for i, x_val in enumerate(categories):
            x_mask = (true == x_val)
            if np.sum(x_mask) > 0:
                for j, w_val in enumerate(categories):
                    w_mask = (mismeasured == w_val)
                    theta[j, i] = np.sum(x_mask & w_mask) / np.sum(x_mask)
        
        return theta, categories

    def _estimate_reclassification_probs(self, theta, categories):
        """
        Estimate reclassification probabilities λ_{x|w} = P(X=x|W=w).
        """
        n_categories = len(categories)
        lambda_probs = np.zeros((n_categories, n_categories))
        numerator_lambda_probs = np.zeros((n_categories, n_categories))
        # Calculate P(X=x)
        pi_hat=np.zeros(n_categories)
        for x in range(n_categories):
            pi_hat[x] = self.val_data[self.val_data[self.mismeasured_var]==categories[x]].shape[0]/self.val_data.shape[0]

        # Calculate \lambda_{x|w} = (\theta_(w|x)*P(X=x))/(Σ_x \theta(w|x)P(X=x))
        # Start with numerator
        for x in range(n_categories):
            for w in range(n_categories):
                numerator_lambda_probs[w, x] = (theta[w, x] * pi_hat[x]) if pi_hat[x] > 0 else 0
        # Then do denominator
        for x in range(n_categories):
            for w in range(n_categories):
                lambda_probs[w, x] = numerator_lambda_probs[w, x] / numerator_lambda_probs[x, :].sum()
                
        return lambda_probs

    def _bootstrap_validation(self, val_data, n_boots):
            """
            Bootstrap the validation data to get distributions of classification probabilities.
            """
            n_val = len(val_data)
            theta_boots = []
            lambda_boots = []

            for _ in tqdm(range(n_boots), desc="Bootstrap validation", leave=True):
                # Bootstrap sample from validation data
                boot_idx = np.random.choice(n_val, size=n_val, replace=True)
                boot_val = val_data.iloc[boot_idx]

                # Estimate classification probabilities
                theta, categories = self._estimate_classification_probs(
                    boot_val[self.mismeasured_var],
                    boot_val[self.true_var]
                )
                
                # Estimate reclassification probabilities
                lambda_probs = self._estimate_reclassification_probs(theta, categories)
                
                theta_boots.append(theta)
                lambda_boots.append(lambda_probs)
                
            return np.array(theta_boots), np.array(lambda_boots)

    def _apply_correction(self, data, lambda_probs, categories):
        """
        Apply reclassification probabilities to convert W to X probabilities.
        """
        n = len(data)
        corrected_w = np.zeros(n)
        
        # Map each W value to its corrected X probability
        for i in range(n):
            w_val = data[self.mismeasured_var].iloc[i]
            w_idx = np.where(categories == w_val)[0][0]
            corrected_w[i] = lambda_probs[1, w_idx]  # P(X=1|W=w)
            
        return corrected_w

    def _bootstrap_regression(self, method, formula, data, val_data, n_boots):
        """
        Perform bootstrap for regression with measurement error correction.
        """
        n = len(data)
        n_val = len(val_data)
        boot_coefs = []
        
        # Use the formula with the mismeasured variable
        formula_to_use = self.y_var+"~"
        for i in range(len(self.x_vars)):
            if i==0:
                formula_to_use = formula_to_use +str(self.x_vars[i])
            if i>0:
                formula_to_use = formula_to_use +'+ '+str(self.x_vars[i])

        for _ in tqdm(range(n_boots), desc=f"Bootstrap {method} regression", leave=True):
            # Bootstrap sample from validation data
            boot_idx = np.random.choice(n_val, size=n_val, replace=True)
            boot_val = val_data.iloc[boot_idx]

            # Estimate classification probabilities
            boot_theta, categories = self._estimate_classification_probs(
                boot_val[self.mismeasured_var],
                boot_val[self.true_var]
            )
            
            # Bootstrap main data
            boot_idx = np.random.choice(n, size=n, replace=True)
            boot_data = data.iloc[boot_idx].copy()

            # Get design matrices
            y, X = dmatrices(formula_to_use, data=boot_data)
            X_arr = np.array(X)
            
            # Apply correction using bootstrap validation estimates
            if method=='OLS':
                # Estimate reclassification probabilities
                lambda_probs = self._estimate_reclassification_probs(boot_theta, categories)
                corrected_w = self._apply_correction(boot_data, lambda_probs, categories)
            
                # Find the column index for the mismeasured variable and replace it
                w_idx = [i for i, name in enumerate(X.design_info.column_names) if self.mismeasured_var in name][0]
                X_arr[:, w_idx] = corrected_w

                # Fit model
                model = sm.OLS(y, X_arr)
                result = model.fit()
                result.params = pd.Series(result.params, index=X.design_info.column_names)
                boot_coefs.append(result.params)
           
            if method=='Logit':
                naive_result = sm.Logit(y, X).fit(disp=0)
                result = self.bi_logit_MLE(naive_result.params,boot_theta)
                boot_coefs.append(result)
            
        return np.array(boot_coefs)
    
    def ols(self, formula, data, val_data=None, bootstrap=False, n_boots=1000):
        """
        Perform OLS regression with validation correction.
        
        Parameters:
        -----------
        formula : str
            Formula in the format "y ~ w || x + z" where w is the mismeasurement of variable x
        data : pandas.DataFrame
            Main dataset with mismeasured variables
        val_data : pandas.DataFrame, optional
            Validation dataset with both true and mismeasured variables
        bootstrap : bool, optional
            Whether to compute bootstrap confidence intervals
        n_boots : int, optional
            Number of bootstrap iterations if bootstrap=True
        """
        self._prepare_data(formula, data, val_data)
        
        # For naive regression, use the part before ||        
        naive_formula = self.y_var+" ~ "
        for i in range(len(self.x_vars)):
            if i==0:
                naive_formula = naive_formula +str(self.x_vars[i])
            if i>0:
                naive_formula = naive_formula +' + '+str(self.x_vars[i])        
        y, X = dmatrices(naive_formula, data=data)
        naive_result = sm.OLS(y, X).fit()
        
        if val_data is None:
            print(f"Naive estimation of {formula}.")
            return naive_result
            
        if not self.is_outcome_mismeasured:
            if bootstrap:
                # Perform bootstrap with validation correction
                coef_boots = self._bootstrap_regression('OLS', formula, data, val_data, n_boots)
                self.bootstrap_results = {
                    'coefficients': coef_boots,
                    'conf_int': np.percentile(coef_boots, [2.5, 97.5], axis=0),
                    'std_err': np.std(coef_boots, axis=0)
                }
                
                # Run point estimate with full validation data
                theta, categories = self._estimate_classification_probs(
                    val_data[self.mismeasured_var],
                    val_data[self.true_var]
                )
                lambda_probs = self._estimate_reclassification_probs(theta, categories)
                
                # Apply correction
                corrected_x = self._apply_correction(self.data, lambda_probs, categories)
                
                # Use the formula with the mismeasured variable
                y, X = dmatrices(naive_formula, data=data)
                X_arr = np.asarray(X)  # Convert to numpy array
                # Find the column index for the mismeasured variable
                w_idx = [i for i, name in enumerate(X.design_info.column_names) if self.mismeasured_var in name][0]
                X_arr[:, w_idx] = corrected_x
                
                # Fit model and create pandas Series with correct index
                model = sm.OLS(y, X_arr)
                result = model.fit()
                
                # Add bootstrap confidence intervals
                my_list = self.x_vars
                for i in range(len(my_list)):
                    if self.mismeasured_var in my_list[i]:
                        my_list[i] = f"corrected_{self.mismeasured_var}"
                my_list.insert(0, 'const') 
                OLS_boot_out = pd.DataFrame({'vars': my_list})
                OLS_boot_out['coeffs'] = pd.DataFrame(result.params)

                conf_int_boot_out = pd.DataFrame(
                    self.bootstrap_results['conf_int'].T,
                    columns=['2.5%', '97.5%']
                )
                conf_int_boot_out['std_err'] = self.bootstrap_results['std_err'].T

                OLS_boot_out[['std_err','[0.025', '0.975]']] = conf_int_boot_out[['std_err','2.5%', '97.5%']]

                print(f"Parameter estimates with bootstrapped confidence intervals are:")
                return OLS_boot_out
            else:
                raise NotImplementedError("Non-bootstrap correction not implemented. Set bootstrap=True.")
        else:
            raise NotImplementedError("Correction for mismeasured outcome not yet implemented.")
        
    def BLL(self, beta, u, x_vars,const, theta):
        return binary_logit_likelihood(beta, u, x_vars, const, theta).neg_logL()

    def bi_logit_MLE(self, init_beta, theta):

        beta=init_beta
        ### run optimization given current guess of parameters
        res = optimize.minimize(self.BLL, 
                                beta,
                                method='L-BFGS-B',
                                tol = 1e-12,
                                args=(self.data[self.y_var],self.data[self.x_vars],True,theta),
                                options={'maxiter':150, 
                                        'gtol': 1e-6, 
                                        'disp': False}
                                )
        
        return res.x

    def logit(self, formula, data, val_data=None, bootstrap=True, n_boots=1000):
        """
        Perform logistic regression with validation correction.
        
        Parameters:
        -----------
        formula : str
            Formula in the format "u || y ~ x + z" for mismeasured outcome (where u is the mismeasured outcome, y is the true outcome)
        data : pandas.DataFrame
            Main dataset with mismeasured variables
        val_data : pandas.DataFrame, optional
            Validation dataset with both true and mismeasured variables
        bootstrap : bool, optional
            Whether to compute bootstrap confidence intervals
        n_boots : int, optional
            Number of bootstrap iterations if bootstrap=True
        """
        self._prepare_data(formula, data, val_data)

        # Formula for naive regression
        naive_formula = self.y_var+" ~ "
        for i in range(len(self.x_vars)):
            if i==0:
                naive_formula = naive_formula +str(self.x_vars[i])
            if i>0:
                naive_formula = naive_formula +' + '+str(self.x_vars[i])              
        y, X = dmatrices(naive_formula, data=data)
        naive_result = sm.Logit(y, X).fit(disp=0)
        
        if val_data is None:
            print(f"Naive estimation of {formula}.")
            return naive_result
        if val_data is not None:
            theta, categories = self._estimate_classification_probs(self.val_data[self.mismeasured_var], self.val_data[self.true_var])
            # run corrected regression for point-estimate
            corr_res = self.bi_logit_MLE(naive_result.params, theta)
            
            if bootstrap:
                # Perform bootstrap with validation correction
                coef_boots = self._bootstrap_regression('Logit', formula, data, val_data, n_boots)
                self.bootstrap_results = {
                    'coefficients': coef_boots,
                    'conf_int': np.percentile(coef_boots, [2.5, 97.5], axis=0),
                    'std_err': np.std(coef_boots, axis=0)
                }
                
                # Add bootstrap confidence intervals
                my_list = self.x_vars
                my_list.insert(0, 'const') 
                logit_boot_out = pd.DataFrame({'vars': my_list})
                logit_boot_out['coeffs'] = pd.DataFrame(corr_res)

                conf_int_boot_out = pd.DataFrame(
                    self.bootstrap_results['conf_int'].T,
                    columns=['2.5%', '97.5%']
                )
                conf_int_boot_out['std_err'] = self.bootstrap_results['std_err'].T

                logit_boot_out[['std_err','[0.025', '0.975]']] = conf_int_boot_out[['std_err','2.5%', '97.5%']]

                print(f"Parameter estimates with bootstrapped confidence intervals are:")
                return logit_boot_out
            else:
                print(f"Parameter point estimates are:")
                print(corr_res)
                return corr_res


    def plot_coefficients(self, naive_result, corrected_result=None, figsize=(10, 6)):
        """Plot comparison of coefficients between naive and corrected models."""
        if corrected_result is None:
            corrected_result = naive_result
        return plot_coefficient_comparison(naive_result, corrected_result, figsize=figsize)

    def plot_bootstrap_distributions(self, param_names=None, figsize=(10, 6)):
        """Plot bootstrap distributions."""
        if self.bootstrap_results is None:
            raise ValueError("No bootstrap results available. Run regression with bootstrap=True first.")
        return plot_bootstrap_distributions(
            coef_boots=self.bootstrap_results['coefficients'],
            param_names=param_names or self.x_vars,
            figsize=figsize
        )


class binary_logit_likelihood:
    def __init__(self, beta, u, x_vars, const, theta):

        ### vars from data
        self.u = u
        self.x_vars = x_vars
        if const:
            if self.x_vars.columns[0]!='const':
                c = np.ones_like(self.u)
                self.x_vars.insert(loc = 0,
                column = 'const',
                value = c)

        self.theta = theta #should be 2x2 numpy array

        ### params to maximize
        self.beta = beta

    def F(self):
        return (1/(1+np.exp(-(self.beta*self.x_vars).sum(axis=1))))

    def p(self):
        return self.F()*(self.theta[1,1]+self.theta[0,0]-1)+1-self.theta[0,0]

    def Psi(self):
        return np.log(self.p()/(1-self.p()))

    def logL(self):
        return self.u*self.Psi() - np.log(1+np.exp(self.Psi()))
    
    def neg_logL(self):
        return -self.logL().sum()/self.u.shape[0]