import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

def plot_coefficient_comparison(naive_result, corrected_result, figsize=(10, 6)):
    """
    Plot a comparison of coefficients between naive and corrected models.
    
    Parameters:
    -----------
    naive_result : RegressionResults
        Results from naive regression (statsmodels results object)
    corrected_result : pandas.DataFrame
        Results from corrected regression with columns: vars, coeffs, std_err, [0.025, 0.975]
    figsize : tuple, optional
        Figure size (width, height)
        
    Returns:
    --------
    matplotlib.figure.Figure
        The generated figure
    """
    # Extract coefficients and confidence intervals from naive regression
    naive_coef = naive_result.params
    naive_ci = naive_result.conf_int()
    
    # Convert to numpy arrays if they aren't already
    if not isinstance(naive_coef, np.ndarray):
        naive_coef = np.array(naive_coef)
    if not isinstance(naive_ci, np.ndarray):
        naive_ci = np.array(naive_ci)
    
    # Extract coefficients and confidence intervals from corrected regression
    corrected_coef = corrected_result['coeffs'].values
    corrected_ci_lower = corrected_result['[0.025'].values
    corrected_ci_upper = corrected_result['0.975]'].values
    
    # Get parameter names
    param_names = corrected_result['vars'].values
    
    # Set up the plot
    fig, ax = plt.subplots(figsize=figsize)
    
    # Number of coefficients
    n_coef = len(param_names)
    y_pos = np.arange(n_coef)
    
    # Plot coefficients and CIs
    ax.errorbar(naive_coef, y_pos - 0.1, 
               xerr=[naive_coef - naive_ci[:, 0], naive_ci[:, 1] - naive_coef],
               fmt='o', label='Naive', color='blue', capsize=5)
    
    ax.errorbar(corrected_coef, y_pos + 0.1,
               xerr=[corrected_coef - corrected_ci_lower, corrected_ci_upper - corrected_coef],
               fmt='o', label='Corrected', color='red', capsize=5)
    
    # Customize plot
    ax.set_yticks(y_pos)
    ax.set_yticklabels(param_names)
    ax.axvline(x=0, color='gray', linestyle='--', alpha=0.5)
    
    ax.set_xlabel('Coefficient Value')
    ax.set_title('Comparison of Naive vs. Corrected Coefficients')
    ax.legend()
    
    plt.tight_layout()
    return fig

def plot_bootstrap_distributions(coef_boots=None, param_names=None, figsize=(10, 6)):
    """
    Plot the distributions of bootstrapped coefficients.
    
    Parameters:
    -----------
    coef_boots : numpy.ndarray
        Bootstrap samples of coefficients (n_boots x n_params)
    param_names : list, optional
        Names of the parameters
    figsize : tuple, optional
        Figure size (width, height)
        
    Returns:
    --------
    matplotlib.figure.Figure
        The generated figure
    """
    if coef_boots is None:
        raise ValueError("Bootstrap coefficients must be provided")
    
    if param_names is None:
        param_names = [f'β{i}' for i in range(coef_boots.shape[1])]
    
    n_params = coef_boots.shape[1]
    fig, axes = plt.subplots(n_params, 1, figsize=figsize)
    if n_params == 1:
        axes = [axes]
    
    for i, (ax, param) in enumerate(zip(axes, param_names)):
        # Convert to numpy array if needed
        data = np.array(coef_boots[:, i])
        
        # Plot bootstrap distribution
        sns.kdeplot(data=data, ax=ax, color='red', fill=True, alpha=0.3,bw_method=0.05)
        
        # Add vertical lines for confidence intervals
        ci = np.percentile(data, [2.5, 97.5])
        ax.axvline(x=ci[0], color='gray', linestyle='--', alpha=0.5, label='95% CI')
        ax.axvline(x=ci[1], color='gray', linestyle='--', alpha=0.5)
        ax.axvline(x=np.mean(data), color='red', linestyle='-', label='Mean')
        
        ax.set_title(f'Bootstrap Distribution of {param}')
        ax.set_xlabel('Coefficient Value')
        ax.legend()
    
    plt.tight_layout()
    return fig 