import numpy as np
import pandas as pd

def parse_formula(formula):
    """
    Parse a formula string to extract variables and determine if outcome is mismeasured.
    
    Parameters:
    -----------
    formula : str
        Formula string in one of two formats:
        - "y ~ w || x + z" (mismeasured predictor, where w is mismeasurement of x)
        - "u || y ~ x + z" (mismeasured outcome, where u is mismeasurement of y)
        
    Returns:
    --------
    dict
        Dictionary containing parsed components:
        - y_var: dependent variable
        - x_vars: list of independent variables
        - mismeasured_var: mismeasured variable
        - true_var: true variable
        - is_outcome_mismeasured: boolean indicating if outcome is mismeasured
    """
    # Split on ~ and strip whitespace from parts
    lhs, rhs = formula.split('~')
    
    if '||' not in formula:
        # No validation correction specified
        return {
            'y_var': lhs.strip(),
            'x_vars': [x.strip() for x in rhs.split('+')],
            'mismeasured_var': None,
            'true_var': None,
            'is_outcome_mismeasured': False
        }
    
    if '||' in rhs:
        # Format: y ~ w || x + z
        # Mismeasured predictor case
        rhs_vars = [x.strip() for x in rhs.split('+')]

        # Identify mismeasured variable and true values
        mismeasured_var = []
        true_var = []
        for i in range(len(rhs_vars)):
            var = rhs_vars[i]
            if '||' in var:
                mismeasured_var, true_var = var.split('||')
                rhs_vars[i] = mismeasured_var

        return {
            'y_var': lhs.strip(),
            'x_vars': rhs_vars,
            'mismeasured_var': mismeasured_var.strip(),
            'true_var': true_var.strip(),
            'is_outcome_mismeasured': False
        }
    if '||' in lhs:
        # Format: u || y ~ x + z
        # Mismeasured outcome case
        parts = [p.strip() for p in lhs.split('||')]
        mismeasured_var = parts[0].strip()
        true_var = parts[1].strip()
        
        return {
            'y_var': mismeasured_var,
            'x_vars': [x.strip() for x in rhs.split('+')],
            'mismeasured_var': mismeasured_var,
            'true_var': true_var,
            'is_outcome_mismeasured': True
        }
    else:
        print('Model misspecified')

def estimate_misclassification_rates(mismeasured, true):
    """
    Estimate misclassification rates from validation data.
    
    Parameters:
    -----------
    mismeasured : array-like
        Mismeasured variable values
    true : array-like
        True variable values
        
    Returns:
    --------
    numpy.ndarray
        Matrix of misclassification probabilities where entry (i,j) is
        P(W=i|X=j)
    """
    # Convert to numpy arrays
    mismeasured = np.array(mismeasured)
    true = np.array(true)
    
    # Get unique categories
    categories = np.unique(np.concatenate([mismeasured, true]))
    n_categories = len(categories)
    
    # Initialize misclassification matrix
    misclass_matrix = np.zeros((n_categories, n_categories))
    
    # Calculate conditional probabilities
    for i, true_val in enumerate(categories):
        true_mask = (true == true_val)
        if np.sum(true_mask) > 0:  # avoid division by zero
            for j, mis_val in enumerate(categories):
                mis_mask = (mismeasured == mis_val)
                misclass_matrix[j, i] = np.sum(true_mask & mis_mask) / np.sum(true_mask)
    
    return misclass_matrix
